"""
Copyright (C) Cross The Road Electronics.  All rights reserved.
License information can be found in CTRE_LICENSE.txt
For support and suggestions contact support@ctr-electronics.com or file
an issue tracker at https://github.com/CrossTheRoadElec/Phoenix-Releases
"""

from phoenix6.status_code import StatusCode
from phoenix6.phoenix_native import Native
import ctypes
from phoenix6.units import *
from phoenix6.sim.chassis_reference import ChassisReference
from phoenix6.sim.device_type import DeviceType

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from phoenix6.hardware.core.core_talon_fx import CoreTalonFX

class TalonFXSimState:
    """
    Creates an object to control the state of a simulated TalonFX.

    Note the recommended method of accessing simulation features is
    to use TalonFX.sim_state.

    :param device: Device to which this simulation state is attached
    :type device: CoreTalonFX
    :param orientation: Orientation of the device relative to the robot chassis
    :type orientation: ChassisReference
    """

    __device_type = DeviceType.PRO_TalonFXType

    def __init__(self, device: 'CoreTalonFX', orientation: ChassisReference = ChassisReference.CounterClockwise_Positive):
        self._id = device.device_id

        self.orientation = orientation
        """
        The orientation of the TalonFX relative to the robot chassis.

        This value should not be changed based on the TalonFX invert.
        Rather, this value should be changed when the mechanical linkage
        between the TalonFX and the robot changes.
        """

    @property
    def last_status_code(self) -> StatusCode:
        """
        Gets the last status code generated by a simulation function.

        Not all functions return a status code but can potentially report errors.
        This function can be used to retrieve those status codes.

        :returns: Last status code generated by a simulation function
        :rtype: StatusCode
        """
        return StatusCode(Native.instance().c_ctre_phoenix6_platform_sim_get_last_error(self.__device_type.value, self._id))

    @property
    def motor_voltage(self) -> volt:
        """
        Gets the simulated output voltage of the motor.
        
        :returns: Voltage applied to the motor in Volts
        :rtype: volt
        """
        value = ctypes.c_double()
        Native.instance().c_ctre_phoenix6_platform_sim_get_physics_value(self.__device_type.value, self._id, ctypes.c_char_p(b"MotorVoltage"), ctypes.byref(value))
        if self.orientation == ChassisReference.Clockwise_Positive:
            value.value = -value.value
        return value.value

    @property
    def torque_current(self) -> ampere:
        """
        Gets the simulated output torque current of the motor.

        Phoenix 6 simulation automatically calculates current.
        
        :returns: Torque current applied to the motor in Amperes
        :rtype: ampere
        """
        value = ctypes.c_double()
        Native.instance().c_ctre_phoenix6_platform_sim_get_physics_value(self.__device_type.value, self._id, ctypes.c_char_p(b"TorqueCurrent"), ctypes.byref(value))
        if self.orientation == ChassisReference.Clockwise_Positive:
            value.value = -value.value
        return value.value

    @property
    def supply_current(self) -> ampere:
        """
        Gets the simulated supply current of the TalonFX.

        Phoenix 6 simulation automatically calculates current.
        
        :returns: Supply current of the TalonFX in Amperes
        :rtype: ampere
        """
        value = ctypes.c_double()
        Native.instance().c_ctre_phoenix6_platform_sim_get_physics_value(self.__device_type.value, self._id, ctypes.c_char_p(b"SupplyCurrent"), ctypes.byref(value))
        return value.value

    def set_supply_voltage(self, volts: volt) -> StatusCode:
        """
        Sets the simulated supply voltage of the TalonFX.

        The minimum allowed supply voltage is 4 V - values below this
        will be promoted to 4 V.

        :param volts: The supply voltage in Volts
        :type volts: volt
        :returns: Status code
        :rtype: StatusCode
        """
        return StatusCode(
            Native.instance().c_ctre_phoenix6_platform_sim_set_physics_input(self.__device_type.value, self._id, ctypes.c_char_p(b"SupplyVoltage"), volts)
        )

    def set_forward_limit(self, closed: bool) -> StatusCode:
        """
        Sets the simulated forward limit switch of the TalonFX.

        :param closed: Whether the limit switch is closed
        :type closed: bool
        :returns: Status code
        :rtype: StatusCode
        """
        return StatusCode(
            Native.instance().c_ctre_phoenix6_platform_sim_set_physics_input(self.__device_type.value, self._id, ctypes.c_char_p(b"ForwardLimit"), 1.0 if closed else 0.0)
        )

    def set_reverse_limit(self, closed: bool) -> StatusCode:
        """
        Sets the simulated reverse limit switch of the TalonFX.

        :param closed: Whether the limit switch is closed
        :type closed: bool
        :returns: Status code
        :rtype: StatusCode
        """
        return StatusCode(
            Native.instance().c_ctre_phoenix6_platform_sim_set_physics_input(self.__device_type.value, self._id, ctypes.c_char_p(b"ReverseLimit"), 1.0 if closed else 0.0)
        )

    def set_raw_rotor_position(self, rotations: rotation) -> StatusCode:
        """
        Sets the simulated raw rotor position of the TalonFX.

        Inputs to this function over time should be continuous, as user calls of
        TalonFX.set_position will be accounted for in the callee.

        The TalonFX integrates this to calculate the true reported rotor position.

        When using the WPI Sim GUI, you will notice a readonly position and settable rawPositionInput.
        The readonly signal is the emulated position which will match self-test in Tuner and the hardware API.
        Changes to rawPositionInput will be integrated into the emulated position.
        This way a simulator can modify the position without overriding hardware API calls for home-ing the sensor.

        :param rotations: The raw position in rotations
        :type rotations: rotation
        :returns: Status code
        :rtype: StatusCode
        """
        if self.orientation == ChassisReference.Clockwise_Positive:
            rotations = -rotations
        return StatusCode(
            Native.instance().c_ctre_phoenix6_platform_sim_set_physics_input(self.__device_type.value, self._id, ctypes.c_char_p(b"RawRotorPosition"), rotations)
        )

    def add_rotor_position(self, drotations: rotation) -> StatusCode:
        """
        Adds to the simulated rotor position of the TalonFX.

        :param drotations: The change in position in rotations
        :type drotations: rotation
        :returns: Status code
        :rtype: StatusCode
        """
        if self.orientation == ChassisReference.Clockwise_Positive:
            drotations = -drotations
        return StatusCode(
            Native.instance().c_ctre_phoenix6_platform_sim_set_physics_input(self.__device_type.value, self._id, ctypes.c_char_p(b"AddRotorPosition"), drotations)
        )

    def set_rotor_velocity(self, rps: rotations_per_second) -> StatusCode:
        """
        Sets the simulated rotor velocity of the TalonFX.

        :param rps: The new velocity in rotations per second
        :type rps: rotations_per_second
        :returns: Status code
        :rtype: StatusCode
        """
        if self.orientation == ChassisReference.Clockwise_Positive:
            rps = -rps
        return StatusCode(
            Native.instance().c_ctre_phoenix6_platform_sim_set_physics_input(self.__device_type.value, self._id, ctypes.c_char_p(b"RotorVelocity"), rps)
        )
